#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <utility.h>

#include "lf.h"
#include "soggetti.h"
#include "sezioni.h"

#include "at2.h"
#include "at2500a.h"

#define ALIAS_LCP	100
#define ALIAS_TCS	200

enum ts { undefined = 0, elenco = 1, etichette = 2 };

// definizione form per etichette
class TEti_idonei_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TEti_idonei_form(): TForm() {};
  TEti_idonei_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TEti_idonei_form() {};
};

class TIdonei_application : public TPrintapp
{
  TRelation*				_rel;
  TMask*      			_msk;
  TEti_idonei_form*	_form_eti;
  TAssoc_array			_categorie;
  TString16 				_tipoidon;
  TString16					_idon1, _idon2, _idon3, _idon4;
  
  int 							_cur1, _cur2, _cur3, _cur4;
  TParagraph_string _cognome_nome;
  TDate 						_data_stampa;
	ts 	      				_tipo_stampa;
	TString16					_codsez, _codsot;

  static bool filter_func_idonei(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  TMask& app_mask() { return *_msk; }
  TIdonei_application() : _data_stampa(TODAY), _cognome_nome("",25) {}
};

HIDDEN inline TIdonei_application& app() { return (TIdonei_application&) main_app(); }

TCursor* TEti_idonei_form::cursor() const { return app().current_cursor(); }

TRelation* TEti_idonei_form::relation() const { return cursor()->relation(); }

void TIdonei_application::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  select_cursor(_cur4); 
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TIdonei_application::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipo_stampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			corpo.reset();
			corpo.update();
  		for (int i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"######"));
			set_row(1,"@7g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@10g#a", &_cognome_nome);
			set_row(1,"@36g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(1,"@47g@ld", FLD(LF_SOGGETTI,SOG_DATAULTID));	// potremmo mettere anche datastato
			set_row(1,"@58g@S", FLD(LF_SOGGETTI,SOG_IDON1));
			set_row(1,"@61g@S", FLD(LF_SOGGETTI,SOG_IDON2));
			set_row(1,"@64g@S", FLD(LF_SOGGETTI,SOG_IDON3));
			set_row(1,"@67g@S", FLD(LF_SOGGETTI,SOG_IDON4));

			set_row(2,"@50g@S", FLD(LF_SOGGETTI,SOG_STATO));
		}
		break;
	}					
}

bool TIdonei_application::filter_func_idonei(const TRelation * rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile(); 
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 catsog = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) catsog);
	}	
	// filtro per idoneit
	if (filtrato)
	{
		TAssoc_array idoneita;
		TString16 idonsog = sog.get(SOG_IDON1);
  	if (idonsog.not_empty() && idonsog.ok())
  		idoneita.add((const char*) idonsog);
		idonsog = sog.get(SOG_IDON2);
  	if (idonsog.not_empty() && idonsog.ok())
  		idoneita.add((const char*) idonsog);
		idonsog = sog.get(SOG_IDON3);
  	if (idonsog.not_empty() && idonsog.ok())
  		idoneita.add((const char*) idonsog);
		idonsog = sog.get(SOG_IDON4);
  	if (idonsog.not_empty() && idonsog.ok())
  		idoneita.add((const char*) idonsog);
		if (idoneita.items() != 0)
		{
  		if (app()._idon1.not_empty() && app()._idon1.ok())
				filtrato = idoneita.is_key((const char*) app()._idon1);
			if (app()._idon2.not_empty() && app()._idon2.ok())
				filtrato = filtrato && idoneita.is_key((const char*) app()._idon2);
			if (app()._idon3.not_empty() && app()._idon3.ok())
				filtrato = filtrato && idoneita.is_key((const char*) app()._idon3);
			if (app()._idon4.not_empty() && app()._idon4.ok())
				filtrato = filtrato && idoneita.is_key((const char*) app()._idon4);
		}				
	}
  return filtrato;
}

bool TIdonei_application::preprocess_page(int file, int counter)
{
	// contatore soggetti stampati 
	// per ora non c'
	if (_tipo_stampa == elenco)
	{
		TString80 nome = current_cursor()->curr().get(SOG_COGNOME);
		nome << " ";
		nome << current_cursor()->curr().get(SOG_NOME);
		_cognome_nome = nome;

		// salto pagina se cambio sezione
		const TString16 codsez = current_cursor()->curr().get(SOG_CODSEZ);		
		const TString16 codsot = current_cursor()->curr().get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				printer().formfeed();
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
	}		
  return TRUE;
}

void TIdonei_application::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if ((densot.ok())&& (densot.not_empty()))
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TIdonei_application::set_print(int m)
{ 
	_tipo_stampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipo_stampa = elenco;
  	//_codsez.spaces(2);
  	//_codsot.spaces(2);
  	_codsez = "**";
  	_codsot = "**";
  	break;
  case F_ETICHETTE:
  	_tipo_stampa = etichette;	
  	break;
  }	
	if (_tipo_stampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
    
    // filtro per sezioni selezionati
	  filtra_sezioni();
                                     
		// filtro per tipo idoneit
    _tipoidon = _msk->get(F_TIPO);
    if (_tipoidon.not_empty() && _tipoidon.ok())
    	current_cursor()->setfilter(format("STATO == \"%s\"",(const char*)_tipoidon));
 		else      
      current_cursor()->setfilter("(TCS->S6 == \"I\") || (TCS->S6 == \"F\")", TRUE);
		
		// filtro per categorie
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty() && catpri.ok())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty() && catsec.ok())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty() && catter.ok())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty() && catqua.ok())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty() && catqui.ok())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty() && catses.ok())
  		_categorie.add((const char*) catses);
		
		// filtro per idoneita
		_idon1 = _msk->get(F_IDON1);  		
		_idon2 = _msk->get(F_IDON2);  		
		_idon3 = _msk->get(F_IDON3);  		
		_idon4 = _msk->get(F_IDON4);  		
		
		current_cursor()->set_filterfunction(filter_func_idonei);

		reset_print();
		crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TIdonei_application::crea_intestazione()
{
  reset_header();
  
  if (_tipo_stampa == elenco)
  {
  	TString sep(132);
  	sep = "ELENCO IDONEI";
  	if ((_tipoidon.ok()) && (_tipoidon.not_empty()))
  	{
  		sep << ": tipo ";
  		sep << _tipoidon;
  		const TString80 dtipo = _msk->get(F_D_TIPO);
  		sep << " ";
  		sep << dtipo;
		}
  	if ((_idon1.ok()) && (_idon1.not_empty()))
  	{
  		sep << " per ";
  		sep << _idon1;
		}
  	if ((_idon2.ok()) && (_idon2.not_empty()))
  	{
  		sep << " ";
  		sep << _idon2;
		}
  	if ((_idon3.ok()) && (_idon3.not_empty()))
  	{
  		sep << " ";
  		sep << _idon3;
		}
  	if ((_idon4.ok()) && (_idon4.not_empty()))
  	{
  		sep << " ";
  		sep << _idon4;
		}
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	const TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@120g%s", (const char*) sep);  
  	
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);

  	set_header(4,"@0gCod.@7gC.@10gCognome e nome@36gNato il@47gData ID@58gTipi idon.");
  	set_header(5,"@47gTipo/Esito");  
  	set_header(6,"@0g------@7g--@10g-------------------------@36g----------@47g----------@58g-- -- -- --");
	}  	
}

bool TIdonei_application::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  // per stampare la denominazione della sezione nell'intestazione
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");

  _cur1 = add_cursor(new TCursor(_rel, "", 1));	//cursore ordinamento per codice
  _cur2 = add_cursor(new TCursor(_rel, "", 2));	//cursore ordinamento per cognome e nome
  _cur3 = add_cursor(new TCursor(_rel, "", 3));	//cursore ordinamento per sezione+sottogruppo+codice
  _cur4 = add_cursor(new TCursor(_rel, "", 4));	//cursore ordinamento per sezione+sottogruppo+cognome e nome

  _msk = new TMask("at2500a");
  
  _form_eti = new TEti_idonei_form("AT_ETSOG");
  
  return TRUE;
}

bool TIdonei_application::user_destroy()
{
  delete _msk;
  delete _rel;
  delete _form_eti;
  return TRUE;
}

int at2500(int argc, char* argv[])
{

  TIdonei_application a;

  a.run(argc, argv, "Elenco idonei");

  return 0;
}
