#include <mask.h>
#include <form.h>
#include <printapp.h>

#include "lf.h"
#include "soggetti.h"
#include "sezioni.h"
#include <comuni.h>

#include "at3.h"
#include "at3400a.h"

#define ALIAS_LCP	100
//#define ALIAS_TCS	200

enum ts { undefined = 0, elenco = 1, 	pagine = 2 };

// definizione form per pagine anagrafiche
class TModificati_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TModificati_form(): TForm() {};
  TModificati_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TModificati_form() {};
};

class TModificati_application : public TPrintapp
{
  static bool filter_func_modificati(const TRelation* rel);

  TRelation*        		_rel;
  TMask*            		_msk;
  TModificati_form*			_form_pag;
  TAssoc_array					_categorie;
  
  int 									_cur1, _cur2, _cur3, _cur4;
  TParagraph_string 		_cognome_nome, _dencom;
  
  TDate 								_data_stampa;
  TDate									_dataini, _datafin;
  ts										_tipo_stampa;
  TString16							_codsez, _codsot;
  
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page (int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(TString16 codsez, TString16 codsot);
  TMask& app_mask() { return *_msk; }

  TModificati_application() : _data_stampa(TODAY), _cognome_nome("",25), _dencom("",50) {}
};

HIDDEN inline TModificati_application& app() { return (TModificati_application&) main_app(); }

TCursor* TModificati_form::cursor() const { return app().current_cursor(); }

TRelation* TModificati_form::relation() const { return cursor()->relation(); }

void TModificati_application::filtra_sezioni()
{
 	TString16 sezini = _msk->get(F_SEZINI);
 	TString16 sotini = _msk->get(F_SOTINI);
 	TString16 sezfin = _msk->get(F_SEZFIN);
 	TString16 sotfin = _msk->get(F_SOTFIN);
  select_cursor(_cur4); 
  TLocalisamfile& fl = current_cursor()->file(LF_SOGGETTI);
  TRectype da(fl.curr());
  TRectype a(fl.curr());   
  da.zero();
  a.zero();
  if ((sezini.not_empty()) && (sezini.ok()))
		da.put(SOG_CODSEZ, sezini);
  if ((sotini.not_empty()) && (sotini.ok()))    	
		da.put(SOG_CODSOT, sotini);
  if ((sezfin.not_empty()) && (sezfin.ok()))
		a.put(SOG_CODSEZ, sezfin);
  if ((sotfin.not_empty()) && (sotfin.ok()))    	
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TModificati_application::set_page(int file, int cnt)
{
	switch (_tipo_stampa)
	{
		case pagine:
		{
			TPrint_section& corpo = _form_pag->get_body();
			corpo.reset();
			corpo.update();
  		for (int i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"######"));
			set_row(1,"@7g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@10g#a", &_cognome_nome);
			set_row(1,"@36g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(1,"@47g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
			set_row(1,"@98g@S", FLD(LF_SOGGETTI,SOG_TELABI));
			set_row(1,"@114g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
			set_row(2,"@47g#a", &_dencom);
			set_row(2,"@114g@S", FLD(LF_SOGGETTI,SOG_TELALT));
		}
		break;					
	}	
}

bool TModificati_application::filter_func_modificati(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile(); 
	//filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		TString16 cat = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	// filtro per data aggiornamento
	if (filtrato && app()._dataini.ok())
  {
  	TDate datamod = sog.get(SOG_DATAULTAGG);
  	filtrato = (datamod > app()._dataini);
	}  	
  return filtrato;
}

bool TModificati_application::preprocess_page(int file, int counter)
{
	// contatore soggetti stampati
	// per ora non c'
	if (_tipo_stampa == elenco)
	{
		TString80 nome = current_cursor()->curr().get(SOG_COGNOME);
		nome << " ";
		nome << current_cursor()->curr().get(SOG_NOME);
		_cognome_nome = nome;
		TString256 localita = "";
		localita << current_cursor()->curr(-ALIAS_LCP).get("S6");
		if (localita.not_empty() && localita.ok())
		{
			localita << " ";
			localita << current_cursor()->curr(-ALIAS_LCP).get("S0");
			localita << " - ";		
		}
		else
		{
			localita = current_cursor()->curr(LF_COMUNI).get(COM_CAPCOM);
			localita << " ";
		}			
		localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita.trim();
		_dencom = localita;

		// salto pagina se cambio sezione
		TString16 codsez = current_cursor()->curr().get(SOG_CODSEZ);		
		TString16 codsot = current_cursor()->curr().get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			//if (!_codsez.blank())
			if (_codsez != "**")
				printer().formfeed();
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
	}		
  return TRUE;
}

void TModificati_application::header_sezione(TString16 codsez, TString16 codsot)
{
	TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString256 intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if ((densot.ok())&& (densot.not_empty()))
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TModificati_application::set_print(int)
{
	_tipo_stampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipo_stampa = elenco;
  	//_codsez.spaces(2);
  	//_codsot.spaces(2);
  	_codsez = "**";
  	_codsez = "**";
  	break;
  case F_PAGINE:
  	_tipo_stampa = pagine;	
  	break;
  }	
	if (_tipo_stampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
    _dataini = _msk->get(F_DATAINI);
		filtra_sezioni();

		// filtro per categorie		
  	TString16 catpri = _msk->get(F_CAT1);
  	TString16 catsec = _msk->get(F_CAT2);
  	TString16 catter = _msk->get(F_CAT3);
  	TString16 catqua = _msk->get(F_CAT4);
  	TString16 catqui = _msk->get(F_CAT5);
  	TString16 catses = _msk->get(F_CAT6);
  	if (catpri.not_empty() && catpri.ok())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty() && catsec.ok())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty() && catter.ok())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty() && catqua.ok())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty() && catqui.ok())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty() && catses.ok())
  		_categorie.add((const char*) catses);
  		
		current_cursor()->set_filterfunction(filter_func_modificati);
		reset_print();
    crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TModificati_application::crea_intestazione()
{
  reset_header();
  
  if (_tipo_stampa == elenco)
  {
  	TString256 sep(132);
  	sep = "ELENCO MODIFICATI";
  	TDate data = _msk->get(F_DATAINI);
  	if (data.ok())
  	{
  		sep << " dal ";
  		sep << data.string();
		}
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@120g%s", (const char*) sep);  
  	
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);
  
  	set_header(4,"@0gCod.@7gC.@10gCognome e nome@36gNato il@47gIndirizzo@98gTelefono abit.@114gTelefono lavoro");
  	set_header(5,"@47gCAP/Localit/Comune/Prov.@114gTelefono altro");
  	set_header(6,"@0g------@7g--@10g-------------------------@36g----------@47g--------------------------------------------------@98g---------------@114g---------------");
	}  	
}

bool TModificati_application::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  //_rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==LOCALITA",1,0,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==COM");
  // per stampare nell'intestazione la denominazione della sezione
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");

  _cur1 = add_cursor(new TCursor(_rel, "", 1));	//cursore ordinamento per codice
  _cur2 = add_cursor(new TCursor(_rel, "", 2));	//cursore ordinamento per cognome e nome
  _cur3 = add_cursor(new TCursor(_rel, "", 3));	//cursore ordinamento per sezione+sottogruppo+codice
  _cur4 = add_cursor(new TCursor(_rel, "", 4));	//cursore ordinamento per sezione+sottogruppo+cognome e nome

	_form_pag = new TModificati_form("AT_PAGIN");
  _msk = new TMask("at3400a");
  return TRUE;
}

bool TModificati_application::user_destroy()
{
  delete _msk;
  delete _rel;
  delete _form_pag;
  return TRUE;
}

int at3400(int argc, char* argv[])
{

  TModificati_application a;

  a.run(argc, argv, "Elenco soggetti modificati");

  return 0;
}
