#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <utility.h>

#include "lf.h"
#include "soggetti.h"
#include "sezioni.h"

#include "at3.h"
#include "at3600a.h"

#define ALIAS_LCP	100
#define ALIAS_TDN	200
#define IDON_SI		"SI"
#define IDON_AF		"AF"

enum ts { undefined = 0, elenco = 1, etichette = 2, cartoline = 3 };

// definizione form per etichette
class TConvBO_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TConvBO_form(): TForm() {};
  TConvBO_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TConvBO_form() {};
};

class TConvBO_application : public TPrintapp
{
  TRelation*				_rel;
  TMask*      			_msk;
  TConvBO_form*			_form_eti;


  
  int 							_cur1, _cur2, _cur3, _cur4;
  TParagraph_string _cognome_nome;
  TDate 						_data_stampa;
  TDate							_dataini, _datafin;
	TString16					_procdon;  
	ts 	      				_tipo_stampa;
	TString16					_codsez, _codsot;


  static bool filter_func_convBO(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  TMask& app_mask() { return *_msk; }
  TConvBO_application() : _data_stampa(TODAY), _cognome_nome("",25) {}
};

HIDDEN inline TConvBO_application& app() { return (TConvBO_application&) main_app(); }

TCursor* TConvBO_form::cursor() const { return app().current_cursor(); }

TRelation* TConvBO_form::relation() const { return cursor()->relation(); }

void TConvBO_application::filtra_sezioni()
{
 	const TString16 sez = _msk->get(F_CODSEZ);
 	const TString16 sot = _msk->get(F_CODSOT);
  select_cursor(_cur4); 
  TRectype da(LF_SOGGETTI);
  if (sez.not_empty())
		da.put(SOG_CODSEZ, sez);
  if (sot.not_empty())
		da.put(SOG_CODSOT, sot);
	current_cursor()->setregion(da, da);
}	

void TConvBO_application::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipo_stampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			corpo.reset();
			corpo.update();
  		for (int i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"######"));
			set_row(1,"@7g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@10g#a", &_cognome_nome);
			set_row(1,"@36g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(1,"@47g@ld", FLD(LF_SOGGETTI,SOG_DATAULTID));	// potremmo mettere anche datastato
			set_row(1,"@58g@S", FLD(LF_SOGGETTI,SOG_IDON1));
			set_row(1,"@61g@S", FLD(LF_SOGGETTI,SOG_IDON2));
			set_row(1,"@64g@S", FLD(LF_SOGGETTI,SOG_IDON3));
			set_row(1,"@67g@S", FLD(LF_SOGGETTI,SOG_IDON4));

			set_row(2,"@50g@S", FLD(LF_SOGGETTI,SOG_STATO));
		}
		break;
	}					
}

bool TConvBO_application::filter_func_convBO(const TRelation* rel)
{   

	bool filtrato = TRUE;
	
	TLocalisamfile& sog = rel->lfile();

/*
	TString16 catsog = sog.get(SOG_CATDON);
	TString 
	if (categoria non di dimissione) && (punto di raccolta = scelto o vuoto)
		filtrato = TRUE;
	else
		filtrato = FALSE;	
	if (filtrato)
	{
		
		// filtro per et
  	TRectype from (sog->curr());
  	TRectype to (sog->curr());
  	from.zero();
  	to.zero();
  	if (app()._dataini.ok())
  	{
  		if (app()._procdon == IDON_SI)
    		from.put(SOG_DATAPROSSI, app()._dataini);
    	else	
    		from.put(SOG_DATAPROSAF, app()._dataini);
    }	
  	if (app()._datafin.ok())
  	{
  		if (app()._procdon == IDON_SI)
    		to.put(SOG_DATAPROSSI, app()._datafin);
			else
    		to.put(SOG_DATAPROSAF, app()._datafin);
		}    		
  	if ((sog->curr() >= from) && (sog->curr() <= to))
    	filtrato = TRUE;
    else
    	filtrato = FALSE;	

	}
	*/	
  return filtrato;
}

bool TConvBO_application::preprocess_page(int file, int counter)
{
	// contatore soggetti stampati 
	// per ora non c'
	if (_tipo_stampa == elenco)
	{
		TString80 nome = current_cursor()->curr().get(SOG_COGNOME);
		nome << " ";
		nome << current_cursor()->curr().get(SOG_NOME);
		_cognome_nome = nome;

		// salto pagina se cambio sezione
		const TString16 codsez = current_cursor()->curr().get(SOG_CODSEZ);		
		const TString16 codsot = current_cursor()->curr().get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				printer().formfeed();
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
	}		
  return TRUE;
}

void TConvBO_application::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if ((densot.ok())&& (densot.not_empty()))
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TConvBO_application::set_print(int m)
{ 
	_tipo_stampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipo_stampa = elenco;
  	_codsez = "**";
  	_codsot = "**";
  	break;
  case F_ETICHETTE:
  	_tipo_stampa = etichette;	
  case F_CARTOLINE:
  	_tipo_stampa = cartoline;	
  	break;
  }	
	if (_tipo_stampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);

/*    
    _dataini = _msk->get(F_DATAINI);
    _datafin = _msk->get(F_DATAFIN);
    _procdon = _msk->get(F_PROCDON1);
*/    
    // filtro per sezioni selezionati
	  filtra_sezioni();

		// filtro per idoneo e non escluso                                     
    current_cursor()->setfilter("((TCS->S6 == \"I\" ) || (TCS->S6 == \"F\")) && (ESCLUSO == \"\")", TRUE);
    
		// filtro per procedura di donazione (controllo che abbia la data)
    /* questo filtro  nella filter_function
    TString16 procdon = _msk->get(F_PROCDON1);
    if (procdon == IDON_SI)
			current_cursor()->setfilter("(DATAPROSSI == \"\")", TRUE);
 		else      
			current_cursor()->setfilter("(DATAPROSAF == \"\")", TRUE);
		*/
		// filtro per categorie
		current_cursor()->set_filterfunction(filter_func_convBO);
		reset_print();
		crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TConvBO_application::crea_intestazione()
{
  reset_header();
  
  if (_tipo_stampa == elenco)
  {
  	TString sep(132);
  	sep = "SCADENZE DI DONAZIONE ";
/*  	TString16 procdon = _msk->get(F_PROCDON1);
 		sep << procdon ;
		TDate data = _msk->get(F_DATAINI);
		if (data.ok())
		{
 			sep << " dal ";
  		sep << data.string();
		}
		data = _msk->get(F_DATAFIN);
		if (data.ok())
		{
 			sep << " al ";
  		sep << data.string();
		}
*/		
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@120g%s", (const char*) sep);  
  	
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);

  	set_header(4,"@0gCod.@7gC.@10gCognome e nome@36gNato il@47gData ID@58gTipi idon.");
  	set_header(5,"@47gTipo/Esito");  
  	set_header(6,"@0g------@7g--@10g-------------------------@36g----------@47g----------@58g-- -- -- --");
	}  	
}

bool TConvBO_application::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
//  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==LOCALITA",1,0,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==COM");
  // per stampare la denominazione della sezione nell'intestazione
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");

  _cur1 = add_cursor(new TCursor(_rel, "", 1));	//cursore ordinamento per codice
  _cur2 = add_cursor(new TCursor(_rel, "", 2));	//cursore ordinamento per cognome e nome
  _cur3 = add_cursor(new TCursor(_rel, "", 3));	//cursore ordinamento per sezione+sottogruppo+codice
  _cur4 = add_cursor(new TCursor(_rel, "", 4));	//cursore ordinamento per sezione+sottogruppo+cognome e nome

  _msk = new TMask("at3600a");
  
  _form_eti = new TConvBO_form("AT_ETSOG");
  
  return TRUE;
}

bool TConvBO_application::user_destroy()
{
  delete _msk;
  delete _rel;
  delete _form_eti;
  return TRUE;
}

int at3600(int argc, char* argv[])
{

  TConvBO_application a;

  a.run(argc, argv, "Convocazioni su punto di raccolta");

  return 0;
}
