#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <utility.h>

#include "lf.h"
#include "soggetti.h"
#include "sezioni.h"

#include "at4.h"
#include "at4100a.h"
#include "at4100b.h"
#include "at4100c.h"

#define ALIAS_LCP	100
#define ALIAS_TCS	200
#define IDON_SI		"SI"
#define IDON_AF		"AF"

enum ts { undefined = 0, elenco = 1, etichette = 2, cartoline = 3 };

// definizione form per etichette
class TScadenze_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TScadenze_form(): TForm() {};
  TScadenze_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TScadenze_form() {};
};

class TScadenze_application : public TPrintapp
{
  TRelation*				_rel;
  TMask*      			_msk;
  TScadenze_form*		_form_eti;
  TScadenze_form*   _form_car;
  TAssoc_array			_categorie;
	  
  int 							_cur1, _cur2, _cur3, _cur4;
  TParagraph_string _cognome_nome;
  TDate 						_data_stampa;
  TDate							_dataini, _datafin, _dataultid;
	bool							_usomodo, _usasez;
	TString16					_giorni;
	TString16					_procdon, _modo, _ab01, _ab02, _ab03, _rh;
	ts 	      				_tipo_stampa;
	TString16					_codsez, _codsot, _orario, _datacart;
	TString80					_invitoper, _data, _presso1, _presso2, _presso3, _presso4;
	TString80					_note, _intest1, _intest2, _intest3, _intest4;

  static bool filter_func_scadenze(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
  ts dati_cartoline();

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void dati_sezione(const TString16 codsez, const TString16 codsot);
  TMask& app_mask() { return *_msk; }
  TScadenze_application() : _data_stampa(TODAY), _cognome_nome("",25) {}
};

HIDDEN inline TScadenze_application& app() { return (TScadenze_application&) main_app(); }

TCursor* TScadenze_form::cursor() const { return app().current_cursor(); }

TRelation* TScadenze_form::relation() const { return cursor()->relation(); }

ts TScadenze_application::dati_cartoline()
{
	TMask msk("at4100b");
 	if (msk.run() == K_ENTER) 
 	{ 
 		_invitoper = msk.get(F_INVITOPER);
 		_data			 = msk.get(F_DATA);
 		_presso1   = msk.get(F_PRESSO1);
 		_presso2   = msk.get(F_PRESSO2);
 		_presso3   = msk.get(F_PRESSO3);
 		_presso4   = msk.get(F_PRESSO4);
 		_datacart  = msk.get(F_DATACART);
 		_orario  	 = msk.get(F_ORARIO);
 		_note   	 = msk.get(F_NOTE);
 		_usasez		 = msk.get_bool(F_USASEZ);
 		return cartoline;
 	}
 	else
 		return undefined;
}

void TScadenze_application::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  select_cursor(_cur4); 
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TScadenze_application::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipo_stampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			corpo.update();
  		for (int i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case cartoline:
		{
			TPrint_section& corpo = _form_car->get_body();
			const TString16 codsez = current_cursor()->curr().get(SOG_CODSEZ);		
			const TString16 codsot = current_cursor()->curr().get(SOG_CODSOT);		
			if ((_codsez!=codsez)||(_codsot!=codsot))
			{
				_codsez = codsez;
				_codsot = codsot;
				dati_sezione(codsez, codsot);
				corpo.reset();
			  TForm_item& intest1 = corpo.find_field(CAR_INTEST1);
			  intest1.set(_intest1);
			  TForm_item& intest2 = corpo.find_field(CAR_INTEST2);
			  intest2.set(_intest2);
			  TForm_item& intest3 = corpo.find_field(CAR_INTEST3);
			  intest3.set(_intest3);
			  TForm_item& intest4 = corpo.find_field(CAR_INTEST4);
			  intest4.set(_intest4);
			  TForm_item& invito = corpo.find_field(CAR_INVITOPER);
	  		invito.set(_invitoper);
			  TForm_item& data = corpo.find_field(CAR_DATA);
	  		data.set(_data);
			  TForm_item& presso1 = corpo.find_field(CAR_PRESSO1);
	  		presso1.set(_presso1);
			  TForm_item& presso2 = corpo.find_field(CAR_PRESSO2);
	  		presso2.set(_presso2);
			  TForm_item& presso3 = corpo.find_field(CAR_PRESSO3);
	  		presso3.set(_presso3);
			  TForm_item& presso4 = corpo.find_field(CAR_PRESSO4);
	  		presso4.set(_presso4);
			  TForm_item& orario = corpo.find_field(CAR_ORARIO);
	  		orario.set(_orario);
			  TForm_item& datacart = corpo.find_field(CAR_DATACART);
	  		datacart.set(_datacart);
			  TForm_item& note = corpo.find_field(CAR_NOTE);
	  		note.set(_note);
			}
			corpo.update();
  		for (int i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"#########"));
			set_row(1,"@10g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@13g@S", FLD(LF_SOGGETTI,SOG_COGNOME));
			set_row(1,"@39g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(1,"@50g@ld", FLD(LF_SOGGETTI,SOG_DATAPROSSI));
			set_row(1,"@61g@ld", FLD(LF_SOGGETTI,SOG_DATAPROSAF));
			set_row(1,"@72g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(1,"@83g@ld", FLD(LF_SOGGETTI,SOG_DATAULTID));
			set_row(1,"@94g@S", FLD(LF_SOGGETTI,SOG_IDON1));
			set_row(1,"@97g@S", FLD(LF_SOGGETTI,SOG_IDON2));
			set_row(1,"@100g@S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			
			set_row(2,"@13g@S", FLD(LF_SOGGETTI,SOG_NOME));
			set_row(2,"@53g@pn", FLD(LF_SOGGETTI,SOG_INTSI,"###"));
			set_row(2,"@64g@pn", FLD(LF_SOGGETTI,SOG_INTAF,"###"));
			set_row(2,"@76g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(2,"@87g@S", FLD(LF_SOGGETTI,SOG_TIPOULTID));
			set_row(2,"@94g@S", FLD(LF_SOGGETTI,SOG_IDON3));
			set_row(2,"@97g@S", FLD(LF_SOGGETTI,SOG_IDON4));
			set_row(2,"@100g@S", FLD(LF_SOGGETTI,SOG_RHANTID));			
		}
		break;
	}					
}

bool TScadenze_application::filter_func_scadenze(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile(); 
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	// filtro per modalita' di convocazione
	if (filtrato && app()._usomodo)
	{
		const TString16 modconv = sog.get(SOG_MODCONV);
		filtrato = (modconv == app()._modo);
	}
	// filtro per data ultima idoneita'
	if (filtrato && app()._dataultid.ok())
	{
		const TDate dataultid = sog.get(SOG_DATAULTID);
		filtrato = (dataultid >= app()._dataultid);
	}
	// filtro per tipizzazione
	if (filtrato && (app()._ab01.not_empty() || app()._ab02.not_empty() || app()._ab03.not_empty()))
	{
		const TString16 ab0 = sog.get(SOG_GRUPPOAB0);
		filtrato = (ab0 == app()._ab01 || ab0 == app()._ab02 || ab0 == app()._ab03);
	}		
	if (filtrato && app()._rh.not_empty())
	{
		const TString16 rh	= sog.get(SOG_RHANTID);
		filtrato = (rh == app()._rh);
	}			
	// filtro per data di prossima donazione
	if (filtrato)
	{   
		TDate datapross(NULLDATE);
		if (app()._procdon == IDON_SI)
			datapross = sog.get(SOG_DATAPROSSI);					
		else
			datapross = sog.get(SOG_DATAPROSAF);
		if (datapross.ok())			
		{
	  	if (app()._dataini.ok())
	  		filtrato = datapross >= app()._dataini;
			if (filtrato)
	  		if (app()._datafin.ok())
	  			filtrato = datapross <= app()._datafin;
		}
		else
			filtrato = FALSE;	  			
	}
	// filtro per giorno preferito
	const TString16& giorni = app()._giorni;
	if (filtrato && giorni != "NNNNNNN")
	{
		bool almenoungiorno = FALSE;
		const TString16 giopredon = sog.get(SOG_GIOPREDON);
		for (int i = 0; i < giorni.len(); i++)
		{
			if (giorni[i] == 'X')
				almenoungiorno = (almenoungiorno || giopredon[i] == 'X');
		}				
		filtrato = almenoungiorno;
	}
  return filtrato;
}

bool TScadenze_application::preprocess_page(int file, int counter)
{
	// contatore soggetti stampati 
	// per ora non c'

	switch (_tipo_stampa)
	{
	case elenco:
	{
		const TString16 codsez = current_cursor()->curr().get(SOG_CODSEZ);		
		const TString16 codsot = current_cursor()->curr().get(SOG_CODSOT);		
		TString80 nome = current_cursor()->curr().get(SOG_COGNOME);
		nome << " ";
		nome << current_cursor()->curr().get(SOG_NOME);
		_cognome_nome = nome;

		// salto pagina se cambio sezione
		//const TString16 codsez = current_cursor()->curr().get(SOG_CODSEZ);		
		//const TString16 codsot = current_cursor()->curr().get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				printer().formfeed();
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
	}
	break;
	}
  return TRUE;
}

void TScadenze_application::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if ((densot.ok())&& (densot.not_empty()))
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

void TScadenze_application::dati_sezione(const TString16 codsez, const TString16 codsot)
{
	_intest1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ1);
	_intest2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ2);
	_intest3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ3);
	_intest4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ4);
	if (_usasez)
	{
		if (_presso1.blank())
		{
			_presso1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE1);
			_presso2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE2);
			_presso3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE3);
			_presso4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE4);
		}			
		if (_orario.blank())
			_orario = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVORA);
		if (_note.blank())
			_note = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVNOTE);
	}
	return;
}

bool TScadenze_application::set_print(int m)
{ 
	_tipo_stampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipo_stampa = elenco;
  break;
  case F_ETICHETTE:
  	_tipo_stampa = etichette;	
	break;  	
  case F_CARTOLINE:
  	_tipo_stampa = dati_cartoline();	
  break;
  }	
	if (_tipo_stampa != undefined)
  {
  	_codsez = "**";
  	_codsot = "**";
    reset_files(); 
    add_file(LF_SOGGETTI);
    _dataini 		= _msk->get(F_DATAINI);
    _datafin 		= _msk->get(F_DATAFIN);
    _dataultid 	= _msk->get(F_DATAULTID);
    _procdon 		= _msk->get(F_PROCDON1);
    _usomodo 		= _msk->get_bool(F_USOMODO);
    _modo 	 		= _msk->get(F_MODO);
    _ab01		 		= _msk->get(F_AB01);
    _ab02		 		= _msk->get(F_AB02);
    _ab03		 		= _msk->get(F_AB03);
    _rh			 		= _msk->get(F_RHANTID);
		_giorni 		= "NNNNNNN";    
    if (_msk->get_bool(F_GPD_LUN))
    	_giorni[0] = 'X';
    if (_msk->get_bool(F_GPD_MAR))
    	_giorni[1] = 'X';
    if (_msk->get_bool(F_GPD_MER))
    	_giorni[2] = 'X';
    if (_msk->get_bool(F_GPD_GIO))
    	_giorni[3] = 'X';
    if (_msk->get_bool(F_GPD_VEN))
    	_giorni[4] = 'X';
    if (_msk->get_bool(F_GPD_SAB))
    	_giorni[5] = 'X';
    if (_msk->get_bool(F_GPD_DOM))
    	_giorni[6] = 'X';
      // filtro per sezioni selezionati
	  filtra_sezioni();
		// filtro per categorie		
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
  	if (catpri.not_empty() && catpri.ok())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty() && catsec.ok())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty() && catter.ok())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty() && catqua.ok())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty() && catqui.ok())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty() && catses.ok())
  		_categorie.add((const char*) catses);
		// filtro per idoneo e non escluso                                     
		//if (_usomodo)
    	//current_cursor()->setfilter(format("(TCS->S6 == \"I\" || TCS->S6 == \"F\") && (ESCLUSO == \"\") && (MODCONV == \"%s\")",(const char*)_modo), TRUE);
		//else
    	current_cursor()->setfilter("((TCS->S6 == \"I\" ) || (TCS->S6 == \"F\")) && (ESCLUSO == \"\")", TRUE);
		// filtro per categorie e altro
		current_cursor()->set_filterfunction(filter_func_scadenze);
		reset_print();
		crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TScadenze_application::crea_intestazione()
{
  reset_header();
  
  if (_tipo_stampa == elenco)
  {
  	TString sep(132);
  	sep = "SCADENZE DI DONAZIONE ";
  	TString16 procdon = _msk->get(F_PROCDON1);
 		sep << procdon ;
		TDate data = _msk->get(F_DATAINI);
		if (data.ok())
		{
 			sep << " dal ";
  		sep << data.string();
		}
		data = _msk->get(F_DATAFIN);
		if (data.ok())
		{
 			sep << " al ";
  		sep << data.string();
		}
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@120g%s", (const char*) sep);  
  	
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);

	  set_header(5,"@0gCod.@10gC.@13gCognome@39gNato il@50gData pr.SI@61gData pr.AF@72gData/Tipo@83gData/Tipo@94gIdon.@100gAB0");
	  set_header(6,"@13gNome@50gInterv. SI@61gInterv. AF@72gult. don.@83gult. idon.@94gper@100gRh");  
	  set_header(7,"@0g---------@10g--@13g-------------------------@39g----------@50g----------@61g----------@72g----------@83g----------@94g-----");
	}  	
}

bool TScadenze_application::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  // per stampare la denominazione della sezione nell'intestazione
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");

  _cur1 = add_cursor(new TCursor(_rel, "", 1));	//cursore ordinamento per codice
  _cur2 = add_cursor(new TCursor(_rel, "", 2));	//cursore ordinamento per cognome e nome
  _cur3 = add_cursor(new TCursor(_rel, "", 3));	//cursore ordinamento per sezione+sottogruppo+codice
  _cur4 = add_cursor(new TCursor(_rel, "", 4));	//cursore ordinamento per sezione+sottogruppo+cognome e nome

  _msk = new TMask("at4100a");
  
  _form_eti = new TScadenze_form("AT_ETSOG");
  _form_car = new TScadenze_form("AT_ETCAR");

  return TRUE;
}

bool TScadenze_application::user_destroy()
{
  delete _msk;
  delete _rel;
  delete _form_eti;
  delete _form_car;
  return TRUE;
}

int at4100(int argc, char* argv[])
{

  TScadenze_application a;

  a.run(argc, argv, "Scadenze di donazione");

  return 0;
}
